@@ -30,7 +30,7 @@
  */
 #include <tcl.h>
 #include <string.h>
-#include "sqlite.h"
+#include "sqlite3.h"
 
 /*
  * If compiled on a machine that doesn't have a 32-bit integer,
@@ -356,30 +356,32 @@ int Md5_Init(Tcl_Interp *interp){
 ** During testing, the special md5sum() aggregate function is available.
 ** inside SQLite.  The following routines implement that function.
 */
-static void md5step(sqlite_func *context, int argc, const char **argv){
+static void md5step(sqlite3_context *context, int argc, sqlite3_value **argv){
   MD5Context *p;
   int i;
   if( argc<1 ) return;
-  p = sqlite_aggregate_context(context, sizeof(*p));
+  p = sqlite3_aggregate_context(context, sizeof(*p));
   if( p==0 ) return;
-  if( sqlite_aggregate_count(context)==1 ){
+  if( sqlite3_aggregate_count(context)==1 ){
     MD5Init(p);
   }
   for(i=0; i<argc; i++){
-    if( argv[i] ){
-      MD5Update(p, (unsigned char*)argv[i], strlen(argv[i]));
+    const char *zData = (char*)sqlite3_value_text(argv[i]);
+    if( zData ){
+      MD5Update(p, (unsigned char*)zData, strlen(zData));
     }
   }
 }
-static void md5finalize(sqlite_func *context){
+static void md5finalize(sqlite3_context *context){
   MD5Context *p;
   unsigned char digest[16];
   char zBuf[33];
-  p = sqlite_aggregate_context(context, sizeof(*p));
+  p = sqlite3_aggregate_context(context, sizeof(*p));
   MD5Final(digest,p);
   DigestToBase16(digest, zBuf);
-  sqlite_set_result_string(context, zBuf, strlen(zBuf));
+  sqlite3_result_text(context, zBuf, -1, SQLITE_TRANSIENT);
 }
-void Md5_Register(sqlite *db){
-  sqlite_create_aggregate(db, "md5sum", -1, md5step, md5finalize, 0);
+void Md5_Register(sqlite3 *db){
+  sqlite3_create_function(db, "md5sum", -1, SQLITE_UTF8, 0, 0, 
+      md5step, md5finalize);
 }
